.small {
    font-size:40%;
}

.smaller {
    font-size:60%;
}

.rst-content dl:not(.docutils) dl dt {
    /* mimick numpydoc’s blockquote style */
    font-weight: normal;
    background: none transparent;
    border-left: none;
    margin: 0 0 12px;
    padding: 3px 0 0;
    font-size: 100%;
}

.rst-content dl:not(.docutils) dl dt code {
    font-size: 100%;
    font-weight: normal;
    background: none transparent;
    border: none;
    padding: 0 2px;
}

.rst-content dl:not(.docutils) dl dt a.reference>code {
    text-decoration: underline;
}