from .scatter import scatter
from .velocity import velocity
from .velocity_embedding import velocity_embedding
from .velocity_embedding_grid import velocity_embedding_grid, velocity_embedding_stream
from .utils import hist
from scanpy.api.pl import paga, paga_compare
